

def validate(segment:str):
    if (not segment.isidentifier()) or (segment == 'static'):
        raise ValueError(f"String '{segment}' is not a valid mount-point segment.")



class MountPoint:
    def __init__(self,path:list[str]):
        (validate(s) for s in path)
        path = [p for p in path if p]
        self.path = path
        self.length = len(path)
        self.base = "/".join(path)+("/" if self.length>0 else '')
        self.sroot = "../"*self.length
    def __eq__(self, other):
        return (isinstance(other, self.__class__)
            and self.path == other.path)
    def __ne__(self, other):
        return not self == other
    def __hash__(self) -> int:
        return hash(self.base)
    
        
class MountDepot:
    def __init__(self):
        self._mounts : set[MountPoint] = set()

    def beg(self,recommendation:str) -> MountPoint:
        # trying as is
        minimal = MountPoint([recommendation])
        if not minimal in self._mounts:
            self._mounts.add(minimal)
            return minimal
        # with digit
        for i in range(2,200):
            digited = MountPoint([recommendation+str(i)])
            if not digited in self._mounts:
                self._mounts.add(digited)
                return digited
        # random
        # for l in range(3):
        #     ran_str = 
        #     ran = MountPoint([ran_str])
        #     if not ran in self._mounts:
        #         self._mounts.add(ran)
        #         return ran

        raise ValueError("Ran out of mount space?")
        
        
        

        






from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .wctree import WCTree

import posixpath,pathlib
from .exceptions import MissingOption,NotIdentifier,EmptyTrickle,UnsupportedFeature,NoChildDates,BuildError
from . import pages
from functools import cached_property,lru_cache
from . import assets
from .specs import getopt
import datetime
from . import gutta_info
from .blobs import Blob

import traceback

class WCNode:
    def __init__(self,parent : WCNode,nid:str,config:dict,tree : WCTree,index_in_parent:int):

        # Basic setup
        self.tree = tree
        self.nid = nid
        if (not self.nid.isidentifier()) or (self.nid[0]=='_'):
            raise NotIdentifier(f"The node id '{self.nid}' is invalid. This is not the title, the node id should be a simple identifier")
        if (self.nid in ["static"]):
            raise NotIdentifier(f"The node id '{self.nid}' is a reserved keyword and it's not valid. Choose something else!")
        self.title = getopt(config,'title', "")        
        self.short_title = getopt(config,'short_title',self.title)
        

        # Parenting and inheritance
        self.parent = parent
        self.level : int = -1
        if parent == None:
            self.level = 0
            self.full_title = self.short_title
            self.prefix = getopt(config,'prefix',"")
            self.npath = self.nid
        else:
            self.level = parent.level + 1
            self.full_title = " - ".join([x for x in [parent.full_title, self.short_title] if x])
            self.npath = parent.npath+"."+self.nid
            try:
                self.prefix = getopt(config,'prefix')
            except MissingOption:
                self.prefix = parent.prefix

        # ===== ENRICHMENT HERE =======
        # Enrich with level data
        hlevel = tree.get_level_by_depth(self.level)
        config = hlevel.enrich(config) 

        # Mounting, Layout and URL
        self.layout = getopt(config,'layout')
        self.do_mount = getopt(config,'mount', self.layout != "trickle")
        if self.do_mount:
            self.recommended_mountpoint = getopt(config,'mountpoint',self.nid if self.parent else '')
            self.mountpoint = self.tree.mounts.beg(self.recommended_mountpoint)
            self.base = self.mountpoint.base
            self.siteroot = self.mountpoint.sroot
            self.assets_path = self.siteroot + "_assets/"
            self.static_path = self.siteroot + "static/"
        else:
            if self.parent:
                self.assets_path = self.parent.assets_path
                self.static_path = self.parent.static_path
                self.siteroot = self.parent.siteroot
            else:
                raise BuildError("The root node must be mounted. Make sure 'root' has mount:yes.")
        
        if (not self.do_mount) or self.do_mount == 'redirect':
            if self.parent:
                self.anchor = self.parent.beg_anchor(self.nid)
            else:
                raise BuildError(f"The root node's 'mount' field is '{self.do_mount}'. The root node must have mount:yes.")



        
        # Description and text
        self.description = Blob(getopt(config,'description',""))
        banner_path = getopt(config,'banner',"")
        self.banner = None
        if banner_path:
            self.banner = assets.ImageAsset.get(banner_path)

        self.show_title = getopt(config,'show_title',True)
        self.list_in_toc = getopt(config,'list_in_toc',True)
        self.infobox = getopt(config,'infobox',False)

        # Images

        pix = assets.parse_pixstring(getopt(config,'pix',""))
        self.pix = [assets.ImageAsset.get(self.asset_pfx(pic)) for pic in pix]
        thumb_path = getopt(config,'thumb',default="")        
        self.thumb = assets.ImageAsset.get(self.asset_pfx(thumb_path))

        try:
            self.feed_img = assets.ImageAsset.get(self.asset_pfx(getopt(config,'feed.img')))
        except MissingOption:
            if self.thumb:
                self.feed_img = self.thumb
            elif len(self.pix)>0:
                self.feed_img = self.pix[0]
            else:
                self.feed_img = None
        
        # Opengraph inherit

        self.og_title = self.full_title
        self.og_site_name = self.tree.webcomic_title
        
        self.og_description = Blob('Webcomics!')
        if self.description:
            self.og_description = self.description
        elif (self.parent) and self.parent.description:
            self.og_description = self.parent.description
        


        # Children
        self.anchor_depot : set[str] = set()
        self.children_specs = getopt(config,'children',{})
        self.children = [WCNode(self,key,spec,tree,i) for i,(key,spec) in enumerate(self.children_specs.items())]
        self.is_leaf = len(self.children) == 0
        if self.is_leaf:
            self._subordinate_leaves = [self]
        else:
            self._subordinate_leaves = [leaf for child in self.children for leaf in child._subordinate_leaves]
        
        # Navigation
        
        self.index_in_parent = index_in_parent
        self.content_is_link = getopt(config,'content_is_link',False)

        # Date
        self.show_date = getopt(config,'show_date',False)
        self.dated = getopt(config,'dated',False)
        if self.dated:
            
            self.date_str = getopt(config,'date',"")
            if (self.date_str) and (not self.is_leaf):
                raise UnsupportedFeature(
                    ("You have a non-leaf node which is manually dated. "
                    "Currently, this is unsupported. Date its children "
                    "instead and the date range will be propagated upwards."))

        self.date : datetime.datetime = None # will be parsed/interpolated later.


        # Opengraph backflow
        try:
            self.og_img = assets.ImageAsset.get(self.asset_pfx(getopt(config,'og.img')))
        except MissingOption:
            if len(self.pix)>0:
                self.og_img = self.pix[0]
            elif self.thumb:
                self.og_img = self.thumb
            elif len(self.children)>0:
                    self.og_img = self.children[0].og_img
            else:
                self.og_img = assets.ImageAsset.get('')
        
        
        
    def asset_pfx(self, assetpath:str)->str:
        if assetpath:
            return self.prefix+assetpath
        else:
            return ""

    @property
    def nodemap(self)->dict:
        nmap = {self.base:self}
        for c in self.children:
            nmap.update(c.nodemap)
        return nmap

    @cached_property
    def _all_children(self)->dict[int,list[WCNode]]:
        return self.children + [grandchild for child in self.children for grandchild in child._all_children]

    @lru_cache
    def _resolve_direction(self,forward:bool):
        if self.layout == 'trickle':
            return self.children[0 if forward else -1]
        else:
            return self
    @lru_cache
    def resolve(self):
        return self._resolve_direction(True)
    @lru_cache
    def resolve_back(self):
        return self._resolve_direction(False)

    def beg_anchor(self,recommendation:str):
        if self.do_mount:
            if not recommendation in self.anchor_depot:
                self.anchor_depot.add(recommendation)
                return recommendation
            for l in range(200):
                digited = recommendation + str(l)
                if not digited in self.anchor_depot:
                    self.anchor_depot.add(digited)
                    return digited
            raise ValueError("Ran out of anchor space??")
        else:
            return self.parent.beg_anchor(recommendation)


    @cached_property
    def unmounted_url(self)->str:
        return self.parent.url + "#" + self.anchor

    @cached_property
    def url(self)->str:
        url = None
        if self.do_mount:
            url = self.base
        if self.layout == 'trickle':
            if len(self.children) == 0:
                raise EmptyTrickle(self.nid, self.title)
            url = self.resolve().url
        elif not self.do_mount:
            url = self.unmounted_url
        return url

    @lru_cache
    def permalink(self,webroot)->str:
        return webroot + self.url

    @cached_property
    def next_node(self):
        if self.parent:
            if len(self.parent.children)>1+self.index_in_parent:
                return self.parent.children[self.index_in_parent+1].resolve()
            else:
                return self.parent.next_node
        else:
            return 'latest'

    @cached_property
    def previous_node(self):
        if self.parent:
            if self.index_in_parent == 0:
                return self.parent.previous_node
            else:
                return self.parent.children[self.index_in_parent-1].resolve_back()
        else:
            return 'first'

    @cached_property
    def _toc_subtree(self)->dict:
        stree = {'show': self.list_in_toc, 'title': self.title,'full_title':self.full_title, 'url': self.url}
        if self.is_leaf:
            stree['leaf'] = True
        else:
            stree['leaf'] = False
            stree['sons'] = [son._toc_subtree for son in self.children]
        return stree

    def _date_limit(self, high:bool)->datetime.datetime:
        if self.date:
            return self.date
        else:
            if self.is_leaf:
                raise NoChildDates(self.nid)
            if high:
                return max([c.date_high for c in self.children])
            else:
                return min([c.date_low for c in self.children])


    @cached_property
    def date_low(self)->datetime.datetime:
        return self._date_limit(False)
    @cached_property
    def date_high(self)->datetime.datetime:
        return self._date_limit(True)
    

    @cached_property
    def date_format(self)->str:
        if self.date:
            return self.date.strftime('%d %B, %Y')
        else:
            lowfmt = self.date_low.strftime('%B %Y')
            highfmt = self.date_high.strftime('%B %Y')
            if lowfmt == highfmt:
                return lowfmt
            else:
                return f"{lowfmt} - {highfmt}"



    @cached_property
    def variables(self)->dict:
        variables = {
            'gutta':gutta_info.gutta_info,
            'title':self.title,    
            'assets':self.assets_path,
            'sroot':self.siteroot,
            'static':self.static_path,
            'full_title':self.full_title,
            'show_title':self.show_title,
            'thumb':self.thumb.vars,
            'head_title':' - '.join([ x for x in [self.tree.webcomic_title,self.full_title] if x]),
            'favicon':self.tree.favicon.vars,
            'permalink': self.permalink(self.tree.webroot),
            
        }
        if self.tree.ganalytics:
            variables['ganalytics'] = self.tree.ganalytics

        if self.do_mount:
            variables['opengraph'] = dict(
                title=self.og_title,
                site_name=self.og_site_name,
                description=self.og_description.plaintext,
                image=dict(
                    url=self.og_img.absolute_url(self.tree.webroot),
                    alt=self.og_description.plaintext,
                    width=self.og_img.image_width,
                    height=self.og_img.image_height
                )
            )

        if self.show_date:
            variables['date'] = self.date_format
        
        if self.description:
            variables['description'] = self.description.html
        if self.banner:
            variables['banner'] = self.banner.vars
        if self.tree.navbar:
            variables['navbar'] = self.tree.navbar
        if self.layout == 'gallery':
            variables['entries'] = []
            for c in self.children:
                entry = {
                    'url': c.url,
                    'title': c.title,
                    'description': c.description.html,
                    'description_plain' : c.description.plaintext,
                    'thumb': c.thumb.vars,
                    
                }
                if c.show_date:
                    entry['date'] = c.date_format
                variables['entries'].append(entry)
                
            
        if self.layout == 'scroll':
            variables['content_is_link'] = self.content_is_link
            variables['pix'] = [pic.vars for pic in self.pix]
            variables['entries'] = [
                
                 {
                    "innerbody":c.innerbody ,
                    "anchor":c.anchor
                }
               
                for c in self.children]
            
            variables['infobox'] = self.infobox
            variables['spinner'] = self.tree.spinner.vars
            
            next = self.next_node
            if next == 'latest':
                variables['is_latest'] = True
            else:
                variables['navigation_destination'] = {
                    'url': next.url,
                    'title': next.title,
                    'full_title':next.full_title
                }

            prev = self.previous_node
            if prev == 'first':
                variables['is_first'] = True
            else:
                variables['back_destination'] = {
                    'url': prev.url,
                    'title': prev.title,
                    'full_title':prev.full_title
                }

        # if self.level == 2:
        #     print(f"==={self.nid}===")
        #     print(variables)
        return variables


    @cached_property
    def body(self)->str:
        vars = {'innerbody':self.innerbody}
        vars.update(self.variables)
        return pages.render_page(self.layout,vars)

    @lru_cache
    def render_page(self,body:str)->str:
        if self.do_mount is True:
            vars =  {'body':body}
            vars.update(self.variables)
            return pages.render_page('base',vars)
        elif self.do_mount == 'redirect':
            vars = dict(redirect=self.siteroot+self.unmounted_url)
            vars.update(self.variables)
            return pages.render_page('redirect',vars)
        elif not self.do_mount:
            raise ValueError(".render_page called on an unmounted node?")
        else:
            raise BuildError(f"Unknown 'mount' value '{self.do_mount}'.")

    @cached_property
    def page(self)->str:
        return self.render_page(self.body)

    @cached_property
    def innerbody(self)->str:
        return pages.render_page(self.layout + "_inner", self.variables)

    def build(self):
        for child in self.children:
            child.build()
        
        if self.do_mount:
            try:
                self.page
            except Exception as e:
                print(f"Error while building node '{self.npath}'")
                traceback.print_exc()

            dir = self.base
            pathlib.Path(dir).mkdir(exist_ok=True)
            page_path = posixpath.join(dir,'index.html')
            with open(page_path,'w') as f:
                f.write(self.page)


import sass

autogen_warning = "/* Autogenerated by gutta - DO NOT EDIT. */\n\n"

def build_style(sources:list[str]) -> str:
    return autogen_warning+sass.compile(string = "\n\n".join(sources),output_style='compressed')


from . import wctree
import click
import os
from distutils.dir_util import copy_tree
import pathlib



def build_website():
    wc = wctree.WCTree()
    wc.build()


def create_website():
    if len(os.listdir(os.getcwd()))>0:
        print(os.listdir(os.getcwd()))
        click.echo("The current folder is not empty! Cannot setup template website.",err=True)
        return False
    
    click.echo("Assembling template webcomic...")
    from_dir = os.path.join(os.path.dirname(__file__),'../sample_comic')
    to_dir = os.getcwd()
    copy_tree(from_dir,to_dir)

    pathlib.Path('_assets').mkdir(exist_ok=True)

    click.echo("Done!")
    return


def clean_website():
    if pathlib.Path('_source/webcomic.yaml').is_file():
        for item in os.listdir(os.getcwd()):
            if not item.startswith("_"):
                print(f"Delete {item}")

        click.echo("Done!")
    else:
        click.echo("Can't find webcomic.yaml! Is this even the right folder?",err=True)


from dateutil import parser
import datetime
from .exceptions import DateParsingFailure,NoDates
import numpy as np

def parse_date(datestr:str)->datetime.datetime:
    datestr = datestr.strip()
    if datestr == 'null':
        return None
    elif datestr == '':
        return None
    else:
        try:
            return parser.parse(datestr)
        except parser.ParserError:
            raise DateParsingFailure(datestr)

def daterpolation(datestrings : list[str])->list[datetime.datetime]:
    if(len(datestrings) == 0):
        return []

    dates = [ None if datestr == 'null' else parse_date(datestr) for datestr in datestrings ]
    if all(d is None for d in dates):
        raise NoDates
    
    indices = []
    tstamps = []

    for i,date in enumerate(dates):
        if date is not None:
            indices.append(i)
            tstamps.append(datetime.datetime.timestamp(date))

    interped_timestamps = np.interp(np.arange(0,len(dates)),indices,tstamps)    
    interped_dates = [datetime.datetime.fromtimestamp(ts) for ts in interped_timestamps]

    return interped_dates



import http.server
import socketserver

def run_webserver():
    PORT = 8069
    Handler = http.server.SimpleHTTPRequestHandler
    httpd = socketserver.TCPServer(('0.0.0.0', PORT), Handler)
    try:
        httpd.serve_forever(poll_interval=0.5)
    except KeyboardInterrupt:
        pass
    return



import pathlib
import json
from functools import cached_property,lru_cache

from . import style,resources,assets,localpaths,dates
from .wcnode import WCNode
from .extras import ExtraPage
from .specs import read_yaml,getopt
from .feeds import Feed
from .exceptions import UnknownHierarchyLevel
from .mountpoints import MountDepot



class HierarchyLevel:
    def __init__(self,spec:dict):
        self.spec = spec
    def enrich(self,otherspec:dict)->dict:
        return dict(list(self.spec.items()) + list(otherspec.items()))


class WCTree:
    def __init__(self):
        config = read_yaml(localpaths.ROOT_SPECFILE_PATH)

        self.webcomic_title = getopt(config,'title',"My Webcomic")
        favicon_src = getopt(config,'favicon','favicon.ico')
        self.favicon = assets.ImageAsset.get(favicon_src)
        spinner_src = getopt(config,'loading_spinner','')
        self.spinner = assets.ImageAsset.get(spinner_src)


        self.mounts = MountDepot()

        hierarchy_spec : dict = getopt(config,'hierarchy')
        self.levels = {hid:HierarchyLevel(spec) for hid,spec in hierarchy_spec.items()}
        self.root = WCNode(None,'root',getopt(config,'root'),self,0)
        
        

        self.leaves = self.root._subordinate_leaves
        self.first = self.leaves[0]
        self.latest = self.leaves[-1]


        navbar_spec = getopt(config,'navbar',False)
        self.navbar = None
        if navbar_spec:
            self.navbar = {
                'brand_icon': assets.ImageAsset.get(getopt(navbar_spec,'brand_icon','')).path,
                'brand_text': getopt(navbar_spec,'brand_text',"")
            }
            self.navbar['entries'] = [
                {'label':entry_label,'dest':self.parse_address(entry_dest)}
                for entry_label, entry_dest in getopt(navbar_spec,'menu',{}).items()
                ]

        self.extra_pages = [ExtraPage(spec,self.root,name) for name,spec in getopt(config,'extras',{}).items()]
        
        self.dated_leaves = [leaf for leaf in self.leaves if leaf.dated]
        self.interpolated_dates = dates.daterpolation( [n.date_str for n in self.dated_leaves] )
        for i,leaf in enumerate(self.dated_leaves):
            leaf.date = self.interpolated_dates[i]

        self.webroot = getopt(config,'webroot',"[You forgot to specify a webroot]")
        self.create_feeds = getopt(config,'feed',False)
        if self.create_feeds:
            feed_entries_level_name = getopt(self.create_feeds,'entries')
            self.feed_entries_depth = self.get_depth_by_level_name(feed_entries_level_name)
            self.feed_entries = self.depth_node_map[self.feed_entries_depth]
            # print(self.all_nodes)
            # print(self.feed_entries_depth)
            # print(self.depth_node_map)

            self.feed = Feed(self.feed_entries,self.webroot,{'title':self.webcomic_title,'description':self.root.description.plaintext})

        self.ganalytics = getopt(config,'google.analytics',False)
        if self.ganalytics:
            self.ganalytics_id = getopt(self.ganalytics,'tracking_id')
        

    @lru_cache
    def get_level_by_depth(self,depth:int)-> HierarchyLevel:
        return list(self.levels.values())[depth]

    @lru_cache
    def get_depth_by_level_name(self,name:str)->int:
        for i,key in enumerate(self.levels.keys()):
            if key == name:
                return i
        raise UnknownHierarchyLevel(name)

    @cached_property
    def all_nodes(self) -> list[WCNode]:
        return [self.root] + self.root._all_children

    @cached_property
    def depth_node_map(self)->dict[int,WCNode]:         
        depth_node_map : dict[int,WCNode] = {}
        for i in range(len(self.levels)):
            depth_node_map[i] = []
        for node in self.all_nodes:
            depth_node_map[node.level].append(node)
        return depth_node_map
        
        
    def parse_address(self,address:str)-> str:
        address = address.strip()
        if address.startswith('gutta.'):
            _, param = address.split('.',maxsplit=1)
            try:
                return {
                    'home':'index.html',
                    'first':self.first.url,
                    'latest':self.latest.url,
                    'rss':'rss.xml',
                    'atom':'atom.xml'
                }[param]
            except KeyError:
                print(f"I don't know what to make of '{address}'...")
                return "404.html"
        elif address.startswith("extras."):
            _,param = address.split('.',maxsplit=1)
            return param+".html"
        else:
            return address

    def build_style(self):
        style_gutta = resources.read_resource('gutta.scss')
        try:
            style_vars = open(localpaths.srcpath('vars.scss'),'r').read()
        except FileNotFoundError:
            style_vars = ""
        try:
            style_overrides = open(localpaths.srcpath('style.scss'),'r').read()
        except FileNotFoundError:
            style_overrides = ""
        compiled_css = style.build_style([style_vars,style_gutta,style_overrides])
        with open(localpaths.staticpath('wc.css'),'w') as f:
            f.write(compiled_css)


    def dump_debug(self):
        debugs = {}
        for node in self.all_nodes:
            debugs[node.npath] = node.variables
        with open('.debug_variables.json','w') as f:
            f.write(json.dumps(debugs)) 


    def build(self):
        pathlib.Path('.nojekyll').touch()
        pathlib.Path(localpaths.STATIC).mkdir(exist_ok=True)
        for resource in ['gutta_logo_icon.svg','gutta_logo24.png']:
            resources.move_resource(resource,localpaths.STATIC)
        
        self.build_style()
        self.root.build()
        [ex.build() for ex in self.extra_pages]
        if self.create_feeds:
            self.feed.build()

        self.dump_debug()


import jinja2
import markdown
import os
from .exceptions import TemplateNotFound

TEMPLATE_DIR = os.path.join(os.path.dirname(__file__),'templates')

environment = jinja2.Environment(
    loader=jinja2.FileSystemLoader(TEMPLATE_DIR),
    trim_blocks=True,
    lstrip_blocks=True
    )

class Template:
    def __init__(self, tname:str, full_page:bool):
        self.tname = tname
        self.full_page = full_page
        self._template = environment.get_template(f"{tname}.jinja")

    def render(self,options:dict)-> str:
        rdr = self._template.render(options)
        if self.full_page:
            rdr = rdr.replace('__GUTTA_LAZY_LOAD','eager',1)
            rdr = rdr.replace('__GUTTA_LAZY_LOAD','lazy')
        return rdr

templates_list = [
    Template('base',True),
    Template('gallery',True),
    Template('gallery_inner',False),
    Template('scroll',True),
    Template('scroll_inner',False),
    Template('trickle_inner',False),
    Template('toc_inner',False),
    Template('extra',True),
    Template('redirect',True)
]

templates = {t.tname : t for t in templates_list}


def render_page(template_name:str,options:dict):
    try:
        template = templates[template_name]
    except KeyError:
        raise TemplateNotFound
    return template.render(options)    
    



import os
from . import localpaths
from click import echo
from functools import cached_property
import imagesize
from .exceptions import BadPath


class Asset:
    _cache = {}
    @classmethod
    def get(cls,path:str):
        try:
            return cls._cache[path]
        except KeyError:
            instance = cls(path)
            cls._cache[path] = instance
            return instance

    def __init__(self,path:str):
        self.path = path
        if path:
            self.local_path = localpaths.assetpath(self.path)
            if not os.path.isfile(self.local_path):
                echo(f"Warning: asset {self.path} was referenced but not found.")
    
    def absolute_url(self,webroot:str):
        return webroot+'_assets/'+self.path

    def __bool__(self):
        return self.path != ''
    
class ImageAsset(Asset):
    def __init__(self, path: str):
        super().__init__(path)
        if path:
            try:
                self.image_width, self.image_height = imagesize.get(self.local_path)
            except FileNotFoundError:
                self.image_width, self.image_height = (64,64)
            except OSError:
                raise BadPath(self.local_path)
        else:
            self.image_width,self.image_height = 1,1

    @cached_property
    def vars(self)->dict:

        return {
            'src':self.path,
            'width':self.image_width,
            'height':self.image_height
            }
        
import re

pixstring_regex = re.compile(r"(.*)\[(\d+)-(\d+)\](.*)")


def parse_pixstring(ps:str)-> list[str]:
    pix = []

    for c in ps.split(','):
        c = c.strip()
        if c=="":
            continue
        
        m = pixstring_regex.match(c)
        if m:
            lows = m.group(2)
            highs = m.group(3)
            digs = len(lows)
            low = int(lows)
            high = int(highs)

            for i in range(low,high+1):
                pix.append(f"{m.group(1)}{str(i).zfill(digs)}{m.group(4)}")
        else:
            pix.append(c)
    return pix




from .exceptions import MissingSpecFile
from . import pages
from .blobs import Blob
from functools import cached_property
from . import localpaths
from .wcnode import WCNode
from .specs import getopt

class ExtraPage:
    def __init__(self,spec:dict, reference_node : WCNode, name:str) -> None:
        self.reference_node = reference_node
        self.from_src = getopt(spec,'from',"")
        self.layout = getopt(spec,'layout',"")
        self.name = name
    @cached_property
    def innerbody(self):
        if self.from_src:
            frompath=localpaths.srcpath(self.from_src)
            try:
                with open(frompath,'r') as f:
                    source = f.read()
            except FileNotFoundError:
                raise MissingSpecFile(frompath)
            
            return Blob(source).html
        elif self.layout == "toc":
            return pages.render_page("toc_inner",{'toc':[self.reference_node._toc_subtree]})
        else:
            print(f"Build failed for extra page '{self.name}', unknown type")
            return pages.mdown("Unknown extra page type - Build failed")
    @cached_property
    def body(self):
        return pages.render_page('extra',dict(content=self.innerbody))

    @cached_property
    def variables(self):
        vars = {}
        for inh in ['assets','banner','sroot','static','navbar','favicon','gutta','ganalytics']:
            try:
                vars[inh]=self.reference_node.variables[inh]
            except KeyError:
                pass
        vars.update({
            'title':self.name,
            "show_title":False,
            'body':self.body,
            'head_title': self.reference_node.tree.webcomic_title + " - "+ self.name
        })
        
        return vars

    @cached_property
    def page(self):
        return pages.render_page('base',self.variables)

    def build(self):
      
        with open(self.name+".html",'w') as f:
            f.write(self.page)